package dev.shadowsoffire.apotheosis.adventure.socket;

import java.util.ListIterator;
import java.util.Iterator;
import java.util.stream.Stream;
import net.minecraft.world.level.storage.loot.LootContext;
import net.minecraft.world.item.ItemStack;
import it.unimi.dsi.fastutil.objects.ObjectArrayList;
import net.minecraft.world.item.enchantment.Enchantment;
import java.util.Map;
import java.util.stream.DoubleStream;
import net.minecraft.server.level.ServerPlayer;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraft.core.BlockPos;
import net.minecraft.world.level.LevelAccessor;
import net.minecraft.world.entity.player.Player;
import com.google.common.collect.UnmodifiableIterator;
import javax.annotation.Nullable;
import java.util.function.Predicate;
import com.google.common.base.Predicates;
import net.minecraft.world.InteractionResult;
import net.minecraft.world.item.context.UseOnContext;
import net.minecraft.world.entity.projectile.AbstractArrow;
import net.minecraft.world.entity.Entity;
import net.minecraft.world.entity.LivingEntity;
import net.minecraft.world.entity.MobType;
import net.minecraft.world.damagesource.DamageSource;
import net.minecraft.world.entity.ai.attributes.AttributeModifier;
import net.minecraft.world.entity.ai.attributes.Attribute;
import java.util.function.BiConsumer;
import net.minecraft.world.entity.EquipmentSlot;
import dev.shadowsoffire.apotheosis.adventure.loot.LootCategory;
import java.util.Collection;
import com.google.common.collect.ImmutableList;
import dev.shadowsoffire.apotheosis.adventure.socket.gem.GemInstance;
import java.util.List;

record SocketedGems(ImmutableList<GemInstance> gems) implements List<GemInstance> {
    public static final SocketedGems EMPTY;
    
    public SocketedGems(final List<GemInstance> gems) {
        this(ImmutableList.copyOf((Collection<? extends GemInstance>)gems));
    }
    
    public void addModifiers(final LootCategory cat, final EquipmentSlot type, final BiConsumer<Attribute, AttributeModifier> map) {
        this.streamValidGems().forEach(inst -> inst.addModifiers(type, map));
    }
    
    public int getDamageProtection(final DamageSource source) {
        return this.streamValidGems().map(inst -> inst.getDamageProtection(source)).reduce(0, Integer::sum);
    }
    
    public float getDamageBonus(final MobType creatureType) {
        return this.streamValidGems().map(inst -> inst.getDamageBonus(creatureType)).reduce(Float::sum).orElse(0.0f);
    }
    
    public void doPostAttack(final LivingEntity user, final Entity target) {
        this.streamValidGems().forEach(inst -> {
            final int old = target.f_19802_;
            target.f_19802_ = 0;
            inst.doPostAttack(user, target);
            target.f_19802_ = old;
        });
    }
    
    public void doPostHurt(final LivingEntity user, final Entity attacker) {
        this.streamValidGems().forEach(inst -> inst.doPostHurt(user, attacker));
    }
    
    public void onArrowFired(final LivingEntity user, final AbstractArrow arrow) {
        this.streamValidGems().forEach(inst -> inst.onArrowFired(user, arrow));
    }
    
    @Nullable
    public InteractionResult onItemUse(final UseOnContext useinst) {
        return this.streamValidGems().map(inst -> inst.onItemUse(useinst)).filter(Predicates.notNull()).findFirst().orElse(null);
    }
    
    public float onShieldBlock(final LivingEntity entity, final DamageSource source, float amount) {
        for (final GemInstance inst : this.gems) {
            if (inst.isValid()) {
                amount = inst.onShieldBlock(entity, source, amount);
            }
        }
        return amount;
    }
    
    public void onBlockBreak(final Player player, final LevelAccessor world, final BlockPos pos, final BlockState state) {
        this.streamValidGems().forEach(inst -> inst.onBlockBreak(player, world, pos, state));
    }
    
    public DoubleStream getDurabilityBonusPercentage(final ServerPlayer user) {
        return this.streamValidGems().mapToDouble(inst -> inst.getDurabilityBonusPercentage(user));
    }
    
    public float onHurt(final DamageSource src, final LivingEntity ent, float amount) {
        for (final GemInstance inst : this.gems) {
            if (inst.isValid()) {
                amount = inst.onHurt(src, ent, amount);
            }
        }
        return amount;
    }
    
    public void getEnchantmentLevels(final Map<Enchantment, Integer> enchantments) {
        this.streamValidGems().forEach(inst -> inst.getEnchantmentLevels(enchantments));
    }
    
    public void modifyLoot(final ObjectArrayList<ItemStack> loot, final LootContext ctx) {
        this.streamValidGems().forEach(inst -> inst.modifyLoot(loot, ctx));
    }
    
    public Stream<GemInstance> streamValidGems() {
        return this.gems.stream().filter(GemInstance::isValid);
    }
    
    @Override
    public int size() {
        return this.gems.size();
    }
    
    @Override
    public boolean isEmpty() {
        return this.gems.isEmpty();
    }
    
    @Override
    public boolean contains(final Object o) {
        return this.gems.contains(o);
    }
    
    @Override
    public Iterator<GemInstance> iterator() {
        return this.gems.iterator();
    }
    
    @Override
    public Object[] toArray() {
        return this.gems.toArray();
    }
    
    @Override
    public <T> T[] toArray(final T[] a) {
        return this.gems.toArray(a);
    }
    
    @Deprecated
    @Override
    public boolean add(final GemInstance e) {
        throw new UnsupportedOperationException();
    }
    
    @Deprecated
    @Override
    public boolean remove(final Object o) {
        throw new UnsupportedOperationException();
    }
    
    @Override
    public boolean containsAll(final Collection<?> c) {
        return this.gems.containsAll(c);
    }
    
    @Deprecated
    @Override
    public boolean addAll(final Collection<? extends GemInstance> c) {
        throw new UnsupportedOperationException();
    }
    
    @Deprecated
    @Override
    public boolean addAll(final int index, final Collection<? extends GemInstance> c) {
        throw new UnsupportedOperationException();
    }
    
    @Override
    public boolean removeAll(final Collection<?> c) {
        return false;
    }
    
    @Deprecated
    @Override
    public boolean retainAll(final Collection<?> c) {
        throw new UnsupportedOperationException();
    }
    
    @Deprecated
    @Override
    public void clear() {
        throw new UnsupportedOperationException();
    }
    
    @Override
    public GemInstance get(final int index) {
        return this.gems.get(index);
    }
    
    @Deprecated
    @Override
    public GemInstance set(final int index, final GemInstance element) {
        throw new UnsupportedOperationException();
    }
    
    @Deprecated
    @Override
    public void add(final int index, final GemInstance element) {
        throw new UnsupportedOperationException();
    }
    
    @Deprecated
    @Override
    public GemInstance remove(final int index) {
        throw new UnsupportedOperationException();
    }
    
    @Override
    public int indexOf(final Object o) {
        return this.gems.indexOf(o);
    }
    
    @Override
    public int lastIndexOf(final Object o) {
        return this.gems.lastIndexOf(o);
    }
    
    @Override
    public ListIterator<GemInstance> listIterator() {
        return this.gems.listIterator();
    }
    
    @Override
    public ListIterator<GemInstance> listIterator(final int index) {
        return this.gems.listIterator(index);
    }
    
    @Override
    public List<GemInstance> subList(final int fromIndex, final int toIndex) {
        return this.gems.subList(fromIndex, toIndex);
    }
    
    static {
        EMPTY = new SocketedGems(ImmutableList.of());
    }
}
